function TpvVector2Helper.GetXXX:TpvVector3;
begin
 result.x:=x;
 result.y:=x;
 result.z:=x;
end;
function TpvVector2Helper.GetXXXX:TpvVector4;
begin
 result.x:=x;
 result.y:=x;
 result.z:=x;
 result.w:=x;
end;
function TpvVector2Helper.GetXXXY:TpvVector4;
begin
 result.x:=x;
 result.y:=x;
 result.z:=x;
 result.w:=y;
end;
function TpvVector2Helper.GetXXY:TpvVector3;
begin
 result.x:=x;
 result.y:=x;
 result.z:=y;
end;
function TpvVector2Helper.GetXXYX:TpvVector4;
begin
 result.x:=x;
 result.y:=x;
 result.z:=y;
 result.w:=x;
end;
function TpvVector2Helper.GetXXYY:TpvVector4;
begin
 result.x:=x;
 result.y:=x;
 result.z:=y;
 result.w:=y;
end;
function TpvVector2Helper.GetXYX:TpvVector3;
begin
 result.x:=x;
 result.y:=y;
 result.z:=x;
end;
function TpvVector2Helper.GetXYXX:TpvVector4;
begin
 result.x:=x;
 result.y:=y;
 result.z:=x;
 result.w:=x;
end;
function TpvVector2Helper.GetXYXY:TpvVector4;
begin
 result.x:=x;
 result.y:=y;
 result.z:=x;
 result.w:=y;
end;
function TpvVector2Helper.GetXYY:TpvVector3;
begin
 result.x:=x;
 result.y:=y;
 result.z:=y;
end;
function TpvVector2Helper.GetXYYX:TpvVector4;
begin
 result.x:=x;
 result.y:=y;
 result.z:=y;
 result.w:=x;
end;
function TpvVector2Helper.GetXYYY:TpvVector4;
begin
 result.x:=x;
 result.y:=y;
 result.z:=y;
 result.w:=y;
end;
function TpvVector2Helper.GetYXX:TpvVector3;
begin
 result.x:=y;
 result.y:=x;
 result.z:=x;
end;
function TpvVector2Helper.GetYXXX:TpvVector4;
begin
 result.x:=y;
 result.y:=x;
 result.z:=x;
 result.w:=x;
end;
function TpvVector2Helper.GetYXXY:TpvVector4;
begin
 result.x:=y;
 result.y:=x;
 result.z:=x;
 result.w:=y;
end;
function TpvVector2Helper.GetYXY:TpvVector3;
begin
 result.x:=y;
 result.y:=x;
 result.z:=y;
end;
function TpvVector2Helper.GetYXYX:TpvVector4;
begin
 result.x:=y;
 result.y:=x;
 result.z:=y;
 result.w:=x;
end;
function TpvVector2Helper.GetYXYY:TpvVector4;
begin
 result.x:=y;
 result.y:=x;
 result.z:=y;
 result.w:=y;
end;
function TpvVector2Helper.GetYYX:TpvVector3;
begin
 result.x:=y;
 result.y:=y;
 result.z:=x;
end;
function TpvVector2Helper.GetYYXX:TpvVector4;
begin
 result.x:=y;
 result.y:=y;
 result.z:=x;
 result.w:=x;
end;
function TpvVector2Helper.GetYYXY:TpvVector4;
begin
 result.x:=y;
 result.y:=y;
 result.z:=x;
 result.w:=y;
end;
function TpvVector2Helper.GetYYY:TpvVector3;
begin
 result.x:=y;
 result.y:=y;
 result.z:=y;
end;
function TpvVector2Helper.GetYYYX:TpvVector4;
begin
 result.x:=y;
 result.y:=y;
 result.z:=y;
 result.w:=x;
end;
function TpvVector2Helper.GetYYYY:TpvVector4;
begin
 result.x:=y;
 result.y:=y;
 result.z:=y;
 result.w:=y;
end;
function TpvVector2Helper.GetRRR:TpvVector3;
begin
 result.r:=r;
 result.g:=r;
 result.b:=r;
end;
function TpvVector2Helper.GetRRRR:TpvVector4;
begin
 result.r:=r;
 result.g:=r;
 result.b:=r;
 result.a:=r;
end;
function TpvVector2Helper.GetRRRG:TpvVector4;
begin
 result.r:=r;
 result.g:=r;
 result.b:=r;
 result.a:=g;
end;
function TpvVector2Helper.GetRRG:TpvVector3;
begin
 result.r:=r;
 result.g:=r;
 result.b:=g;
end;
function TpvVector2Helper.GetRRGR:TpvVector4;
begin
 result.r:=r;
 result.g:=r;
 result.b:=g;
 result.a:=r;
end;
function TpvVector2Helper.GetRRGG:TpvVector4;
begin
 result.r:=r;
 result.g:=r;
 result.b:=g;
 result.a:=g;
end;
function TpvVector2Helper.GetRGR:TpvVector3;
begin
 result.r:=r;
 result.g:=g;
 result.b:=r;
end;
function TpvVector2Helper.GetRGRR:TpvVector4;
begin
 result.r:=r;
 result.g:=g;
 result.b:=r;
 result.a:=r;
end;
function TpvVector2Helper.GetRGRG:TpvVector4;
begin
 result.r:=r;
 result.g:=g;
 result.b:=r;
 result.a:=g;
end;
function TpvVector2Helper.GetRGG:TpvVector3;
begin
 result.r:=r;
 result.g:=g;
 result.b:=g;
end;
function TpvVector2Helper.GetRGGR:TpvVector4;
begin
 result.r:=r;
 result.g:=g;
 result.b:=g;
 result.a:=r;
end;
function TpvVector2Helper.GetRGGG:TpvVector4;
begin
 result.r:=r;
 result.g:=g;
 result.b:=g;
 result.a:=g;
end;
function TpvVector2Helper.GetGRR:TpvVector3;
begin
 result.r:=g;
 result.g:=r;
 result.b:=r;
end;
function TpvVector2Helper.GetGRRR:TpvVector4;
begin
 result.r:=g;
 result.g:=r;
 result.b:=r;
 result.a:=r;
end;
function TpvVector2Helper.GetGRRG:TpvVector4;
begin
 result.r:=g;
 result.g:=r;
 result.b:=r;
 result.a:=g;
end;
function TpvVector2Helper.GetGRG:TpvVector3;
begin
 result.r:=g;
 result.g:=r;
 result.b:=g;
end;
function TpvVector2Helper.GetGRGR:TpvVector4;
begin
 result.r:=g;
 result.g:=r;
 result.b:=g;
 result.a:=r;
end;
function TpvVector2Helper.GetGRGG:TpvVector4;
begin
 result.r:=g;
 result.g:=r;
 result.b:=g;
 result.a:=g;
end;
function TpvVector2Helper.GetGGR:TpvVector3;
begin
 result.r:=g;
 result.g:=g;
 result.b:=r;
end;
function TpvVector2Helper.GetGGRR:TpvVector4;
begin
 result.r:=g;
 result.g:=g;
 result.b:=r;
 result.a:=r;
end;
function TpvVector2Helper.GetGGRG:TpvVector4;
begin
 result.r:=g;
 result.g:=g;
 result.b:=r;
 result.a:=g;
end;
function TpvVector2Helper.GetGGG:TpvVector3;
begin
 result.r:=g;
 result.g:=g;
 result.b:=g;
end;
function TpvVector2Helper.GetGGGR:TpvVector4;
begin
 result.r:=g;
 result.g:=g;
 result.b:=g;
 result.a:=r;
end;
function TpvVector2Helper.GetGGGG:TpvVector4;
begin
 result.r:=g;
 result.g:=g;
 result.b:=g;
 result.a:=g;
end;
function TpvVector2Helper.GetSSS:TpvVector3;
begin
 result.s:=s;
 result.t:=s;
 result.p:=s;
end;
function TpvVector2Helper.GetSSSS:TpvVector4;
begin
 result.s:=s;
 result.t:=s;
 result.p:=s;
 result.q:=s;
end;
function TpvVector2Helper.GetSSST:TpvVector4;
begin
 result.s:=s;
 result.t:=s;
 result.p:=s;
 result.q:=t;
end;
function TpvVector2Helper.GetSST:TpvVector3;
begin
 result.s:=s;
 result.t:=s;
 result.p:=t;
end;
function TpvVector2Helper.GetSSTS:TpvVector4;
begin
 result.s:=s;
 result.t:=s;
 result.p:=t;
 result.q:=s;
end;
function TpvVector2Helper.GetSSTT:TpvVector4;
begin
 result.s:=s;
 result.t:=s;
 result.p:=t;
 result.q:=t;
end;
function TpvVector2Helper.GetSTS:TpvVector3;
begin
 result.s:=s;
 result.t:=t;
 result.p:=s;
end;
function TpvVector2Helper.GetSTSS:TpvVector4;
begin
 result.s:=s;
 result.t:=t;
 result.p:=s;
 result.q:=s;
end;
function TpvVector2Helper.GetSTST:TpvVector4;
begin
 result.s:=s;
 result.t:=t;
 result.p:=s;
 result.q:=t;
end;
function TpvVector2Helper.GetSTT:TpvVector3;
begin
 result.s:=s;
 result.t:=t;
 result.p:=t;
end;
function TpvVector2Helper.GetSTTS:TpvVector4;
begin
 result.s:=s;
 result.t:=t;
 result.p:=t;
 result.q:=s;
end;
function TpvVector2Helper.GetSTTT:TpvVector4;
begin
 result.s:=s;
 result.t:=t;
 result.p:=t;
 result.q:=t;
end;
function TpvVector2Helper.GetTSS:TpvVector3;
begin
 result.s:=t;
 result.t:=s;
 result.p:=s;
end;
function TpvVector2Helper.GetTSSS:TpvVector4;
begin
 result.s:=t;
 result.t:=s;
 result.p:=s;
 result.q:=s;
end;
function TpvVector2Helper.GetTSST:TpvVector4;
begin
 result.s:=t;
 result.t:=s;
 result.p:=s;
 result.q:=t;
end;
function TpvVector2Helper.GetTST:TpvVector3;
begin
 result.s:=t;
 result.t:=s;
 result.p:=t;
end;
function TpvVector2Helper.GetTSTS:TpvVector4;
begin
 result.s:=t;
 result.t:=s;
 result.p:=t;
 result.q:=s;
end;
function TpvVector2Helper.GetTSTT:TpvVector4;
begin
 result.s:=t;
 result.t:=s;
 result.p:=t;
 result.q:=t;
end;
function TpvVector2Helper.GetTTS:TpvVector3;
begin
 result.s:=t;
 result.t:=t;
 result.p:=s;
end;
function TpvVector2Helper.GetTTSS:TpvVector4;
begin
 result.s:=t;
 result.t:=t;
 result.p:=s;
 result.q:=s;
end;
function TpvVector2Helper.GetTTST:TpvVector4;
begin
 result.s:=t;
 result.t:=t;
 result.p:=s;
 result.q:=t;
end;
function TpvVector2Helper.GetTTT:TpvVector3;
begin
 result.s:=t;
 result.t:=t;
 result.p:=t;
end;
function TpvVector2Helper.GetTTTS:TpvVector4;
begin
 result.s:=t;
 result.t:=t;
 result.p:=t;
 result.q:=s;
end;
function TpvVector2Helper.GetTTTT:TpvVector4;
begin
 result.s:=t;
 result.t:=t;
 result.p:=t;
 result.q:=t;
end;
